

from typing import List
from pydantic import BaseModel, ConfigDict
from torch import Tensor
from deep_sort_reid.types.coords import CoordinatesXYXY


class Detection(BaseModel):
    model_config = ConfigDict(from_attributes=True,
                              arbitrary_types_allowed=True)

    cls: List[int]  # Classes separated by commas
    coords: CoordinatesXYXY
    confidence: float
    feature: Tensor | None



from pydantic import BaseModel, ConfigDict


class CoordinatesXYAH(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    center_x: float
    center_y: float
    aspect_ratio: float
    height: float


class CoordinatesXYXY(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    start_x: float
    start_y: float
    end_x: float
    end_y: float





from typing import List

from torch import Tensor
import torch
from deep_sort_reid.constants.tracker import GATING_THRESHOLD
from deep_sort_reid.storage.CacheStorage import CacheStorage
from deep_sort_reid.models.motion.KalmanFilter import KalmanFilter
from deep_sort_reid.tracker.Track import Track
from deep_sort_reid.types.coords import CoordinatesXYAH
from deep_sort_reid.types.detection import Detection
from deep_sort_reid.types.metric import MetricType
from deep_sort_reid.metric.Metric import Metric
from deep_sort_reid.utils.box_methods import from_xyah_to_tensor, from_xyxy_to_xyah


class GatedMetric(Metric):

    def __init__(self, metric_type: MetricType, max_distance: float, cache_storage: CacheStorage, kf: KalmanFilter, gated: bool = True):
        self.metric_type: MetricType = metric_type
        self.max_distance = max_distance
        self.cache_storage: CacheStorage = cache_storage
        self.kf: KalmanFilter = kf
        self.gated: bool = gated
        # Distance usage -> f.e do we want min over s

    def __distance(self,
                   tracks: List[Track],
                   detections: List[Detection]) -> Tensor:

        cost_matrix = torch.zeros((len(tracks)), len(detections))

        if self.metric_type == 'iou':
            cost_matrix += torch.inf
            return cost_matrix

        features = []
        measurements: List[Tensor] = []
        for detection in detections:
            # If features are disabled
            # Might have to replace this with torch.zeros and use
            # features dim from somewhere
            measurements.append(from_xyah_to_tensor(
                from_xyxy_to_xyah(detection.coords)))
            if detection.feature is None:
                features.append([])
                continue

            features.append(detection.feature)

        for track_idx, track in enumerate(tracks):
            if self.metric_type == 'cosine':
                cost_matrix[track_idx, :] = self.cosine_distance(
                    self.cache_storage[track.track_id], features)
            elif self.metric_type == 'euclidean':
                cost_matrix[track_idx, :] = self.euclidean_distance(
                    self.cache_storage[track.track_id], features)

        cost_matrix = self.__gated(cost_matrix, tracks, measurements)
        return cost_matrix

    def __gated(self, cost_matrix: Tensor, tracks: List[Track], measurements: List[Tensor]):
        gating_threshold = GATING_THRESHOLD
        measurements_tensor = torch.stack(measurements)

        for track_idx, track in enumerate(tracks):
            gating_distance = self.kf.gating_distance(
                track.state_mean, track.state_covariance, measurements_tensor)

            cost_matrix[track_idx, gating_distance >
                        gating_threshold] = torch.inf

        return cost_matrix

    @staticmethod
    def cosine_distance(samples: List[Tensor], features: List[Tensor]):
        samples_mat = torch.stack(samples)
        features_mat = torch.stack(features)

        A = samples_mat / torch.linalg.norm(samples_mat, dim=1, keepdim=True)
        B = features_mat / torch.linalg.norm(features_mat, dim=1, keepdim=True)

        distances = torch.min((1 - (A @ B.T)), dim=0).values

        return distances

    @staticmethod
    def euclidean_distance(samples: List[Tensor], features: List[Tensor]):
        samples_mat = torch.stack(samples)
        features_mat = torch.stack(features)

        A = samples_mat.unsqueeze(1)
        B = features_mat.unsqueeze(0)

        diff = A - B

        distances = torch.sqrt(torch.sum(torch.square(diff), dim=2))
        distances = torch.min(distances, dim=0).values

        return distances

    @staticmethod
    def similarity(A: Tensor, B: Tensor):

        A_norm = A / torch.linalg.norm(A)
        B_norm = B / torch.linalg.norm(B)

        return A_norm@B_norm

    def __call__(self,
                 tracks: List[Track],
                 detections: List[Detection],
                 ):

        return self.__distance(tracks, detections)





import torch
import torch.backends
import torch.backends.mps


def get_device():
    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available():
        device = "mps"

    return device





from typing import Tuple

from torch import Tensor
import torch
from deep_sort_reid.constants.tracker import AR_POS_STD_DEFAULT, AR_PROJ_STD_DEFAULT, AR_VEL_STD_DEFAULT, MEASUREMENT_SPACE_DIM, POS_WEIGHT_DEFAULT, POS_WEIGHT_FACTOR_INIT, STATE_MEAN_HEIGHT_IDX, VEL_WEIGHT_DEFAULT, VEL_WEIGHT_FACTOR_INIT
from deep_sort_reid.types.coords import CoordinatesXYAH

from deep_sort_reid.utils.box_methods import from_xyah_to_tensor


class KalmanFilter():

    def __init__(self):
        dim = MEASUREMENT_SPACE_DIM
        dt = 1

        # Projects mean and cov to measurement space, discarding velocity information
        self.projection_matrix = torch.eye(dim, 2*dim)

        # Transforms mean and cov using velocity information
        self.transformation_matrix = torch.eye(2 * dim, 2 * dim)
        for i in range(dim):
            self.transformation_matrix[i, dim + i] = dt

    def initiate(self, detection_coords: CoordinatesXYAH,
                 pos_weight=POS_WEIGHT_DEFAULT,
                 vel_weight=VEL_WEIGHT_DEFAULT) -> Tuple[Tensor, Tensor]:

        assert isinstance(detection_coords, CoordinatesXYAH), Exception(
            "detection_coords invalid type")

        measurement = from_xyah_to_tensor(detection_coords)

        state_mean_pos = measurement
        state_mean_vel = torch.zeros_like(state_mean_pos)
        state_mean = torch.cat((state_mean_pos, state_mean_vel))

        pos_weight *= POS_WEIGHT_FACTOR_INIT
        vel_weight *= VEL_WEIGHT_FACTOR_INIT
        state_std_pos = Tensor([
            pos_weight * detection_coords.height,
            pos_weight * detection_coords.height,
            AR_POS_STD_DEFAULT,
            pos_weight * detection_coords.height
        ])
        state_std_vel = Tensor([
            vel_weight * detection_coords.height,
            vel_weight * detection_coords.height,
            AR_VEL_STD_DEFAULT,
            vel_weight * detection_coords.height
        ])

        state_std = torch.cat((state_std_pos, state_std_vel))
        state_covariance = torch.diag(torch.square(state_std))

        return (state_mean, state_covariance)

    def predict(self, state_mean: Tensor, state_covariance: Tensor,
                pos_weight=POS_WEIGHT_DEFAULT,
                vel_weight=VEL_WEIGHT_DEFAULT):

        dim = MEASUREMENT_SPACE_DIM
        assert state_mean.shape[0] == (
            2*dim), "Incorrect state_mean shape"
        assert state_covariance.shape == (
            2*dim, 2*dim), "Incorrect state_covariance shape"

        """
        We introduce noise proportional to the height of the `state_mean`, 
        similarly to how the filter was initialized. 

        This noise accounts for increased uncertainty in the position and velocity 
        based on object size (height), and it helps adjust the predicted state covariance matrix.

        Higher uncertainty is modeled for larger objects, which affects the filter update step.
        """
        state_mean_height = state_mean[STATE_MEAN_HEIGHT_IDX]
        state_std_pos_noise = Tensor([
            pos_weight * state_mean_height,
            pos_weight * state_mean_height,
            AR_POS_STD_DEFAULT,
            pos_weight * state_mean_height
        ])
        state_std_vel_noise = Tensor([
            vel_weight * state_mean_height,
            vel_weight * state_mean_height,
            AR_VEL_STD_DEFAULT,
            vel_weight * state_mean_height
        ])

        state_std_noise = torch.cat(
            (state_std_pos_noise, state_std_vel_noise))
        state_covariance_noise = torch.diag(torch.square(state_std_noise))

        predicted_state_mean = self.transformation_matrix @ state_mean

        predicted_state_covariance = (self.transformation_matrix
                                      @ state_covariance
                                      @ self.transformation_matrix.T) + state_covariance_noise

        return (predicted_state_mean, predicted_state_covariance)

    def __project(self, state_mean, state_covariance, pos_weight=POS_WEIGHT_DEFAULT):

        state_mean_height = state_mean[STATE_MEAN_HEIGHT_IDX]
        measurement_std_noise = Tensor([
            pos_weight * state_mean_height,
            pos_weight * state_mean_height,
            AR_PROJ_STD_DEFAULT,
            pos_weight * state_mean_height
        ])

        measurement_covariance_noise = torch.diag(
            torch.square(measurement_std_noise))

        # Project the state mean to the measurement space
        measurement_mean = self.projection_matrix @ state_mean

        # Project the state covariance to the measurement space, adding further noise
        measurement_covariance = (self.projection_matrix
                                  @ state_covariance
                                  @ self.projection_matrix.T) + measurement_covariance_noise

        return measurement_mean, measurement_covariance

    def update(self, state_mean: Tensor, state_covariance: Tensor,
               detection_coords: CoordinatesXYAH,
               pos_weight=POS_WEIGHT_DEFAULT,
               vel_weight=VEL_WEIGHT_DEFAULT):

        measurement_mean, measurement_covariance = self.__project(
            state_mean, state_covariance, pos_weight)

        """
        Calculation of Kalman Gain

        Let 
        H = projection_matrix
        P = state_covariance
        R = measurement_covariance_noise
        K = kalman_gain
        A = L L.T = (H P H.T + R)
        B = P H.T
    
        Then
        A K = B

        And 
        K = A^-1 B = (L.T)^-1 (L)^-1 B
        """
        L = torch.linalg.cholesky(measurement_covariance)
        B = state_covariance @ self.projection_matrix.T
        K = torch.cholesky_solve(B.T, L).T

        measurement = from_xyah_to_tensor(detection_coords)
        residuals = measurement - measurement_mean

        """
            Updating the predictions and uncertainties using the Kalman Gain. 
            To simplify, a low K would mean higher trust in the prediction,
            while a high K would mean higher trust in the measurement.
        """

        new_mean = state_mean + (residuals@K.T)
        new_covariance = state_covariance - (K@measurement_covariance@K.T)

        return new_mean, new_covariance

    def gating_distance(self, state_mean, state_covariance, measurements_tensor: Tensor):
        """
        Calculates the Mahalanobis distance, 
        which can be thought off as residuals normalized 
        using the state uncertainty (covariance), in turn yielding
        a better measurement as a distance between the state mean and the measurements
        """
        measurement_mean, measurement_covariance = self.__project(
            state_mean, state_covariance)

        """
        Solves for K in
        LK = B
        """
        B = measurements_tensor - measurement_mean
        L = torch.linalg.cholesky(measurement_covariance)
        K = torch.linalg.solve_triangular(L, B.T, upper=False)

        squared_maha = torch.sum(K * K, dim=0)
        return squared_maha





from typing import Dict, List, Literal
import numpy as np
from torch import Tensor
import torch

from deep_sort_reid.types.storage import CacheGetStrategy
from deep_sort_reid.types.tracker import TrackID


class CacheStorage():
    """
    Cache of the feature samples used for re-matching tracks based on appearance.
    """

    samples: Dict[TrackID, List[Tensor]] = {}

    def __init__(self, max_samples_per_track: int):
        self.max_samples_per_track = max_samples_per_track

    def add_sample(self, track_id: TrackID, feature: Tensor):
        # We may want to store samples in different ways, f.e as mean features
        if track_id in self.samples:
            if len(self.samples[track_id]) > self.max_samples_per_track:
                self.samples[track_id] = self.samples[track_id][1:]

            self.samples[track_id].append(feature)

        else:
            self.samples[track_id] = [feature]

    def get(self, key: TrackID, strategy: CacheGetStrategy):
        if strategy == 'all':
            return self.samples[key]
        elif strategy == 'random':
            rand_idx = np.random.randint(0, len(self.samples[key])-1)
            return [self.samples[key][rand_idx]]
        elif strategy == 'mean':
            stacked_tensors = torch.stack(self.samples[key])
            return [torch.mean(stacked_tensors, dim=0)]

    def __getitem__(self, key):
        return self.samples[key]




from typing import List

import cv2
import numpy as np
from torch import Tensor
import torch
from deep_sort_reid.types.detection import Detection
from torchvision.models import resnet50, ResNet50_Weights
from PIL import Image


def extract_features_resnet(file_input_path: str,
                            frames_detections: List[List[Detection]],
                            verbose=False
                            ) -> List[List[Tensor]]:

    weights = ResNet50_Weights.DEFAULT
    model = resnet50(weights=weights)
    model.eval()
    feature_extractor = torch.nn.Sequential(*list(model.children())[:-1])
    preprocess = weights.transforms()

    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available():
        device = "mps"

    feature_extractor.to(device)

    cap = cv2.VideoCapture(file_input_path)

    if not cap.isOpened():
        print("Can't open video file")

    frame_idx = 0
    frames_features: List[List[Tensor]] = []
    while cap.isOpened():
        ret, frame = cap.read()
        if not ret:
            break

        if verbose:
            print("EXTRACTING FEATURES FRAME IDX ", frame_idx)
        detections = frames_detections[frame_idx]

        batch = []
        for detection in detections:
            # Images are indexed as (H, W, C)
            cropped_img = frame[int(detection.coords.start_y):int(detection.coords.end_y),
                                int(detection.coords.start_x):int(detection.coords.end_x)]

            cropped_img = Image.fromarray(cropped_img)
            model_input = (preprocess(
                cropped_img).unsqueeze(0)).to(device)
            batch.append(model_input)

        features = []
        if batch:
            batch = torch.cat(batch)
            object_features: Tensor = (feature_extractor(batch))

            for feature in object_features:
                features.append((feature.detach()).flatten())

        frames_features.append(features)
        frame_idx += 1

    return frames_features





from typing import List, Sequence
from torch import Tensor
from deep_sort_reid.enums.tracker import TrackState
from deep_sort_reid.models.motion.KalmanFilter import KalmanFilter
from deep_sort_reid.types.coords import CoordinatesXYXY
from deep_sort_reid.types.detection import Detection

from deep_sort_reid.types.tracker import TrackID
from deep_sort_reid.utils.box_methods import from_xyxy_to_xyah


class Track():

    def __init__(self, state_mean,
                 state_covariance,
                 track_id,
                 hits_until_confirm,
                 max_since_update,
                 feature=None,
                 cls=[],
                 ):

        self.state_mean: Tensor = state_mean
        self.state_covariance: Tensor = state_covariance
        self.track_id: TrackID = track_id
        self.hits: int = 1
        self.time_since_update: int = 0
        self.hits_until_confirm: int = hits_until_confirm
        self.max_since_update: int = max_since_update
        self.cls: List[int] = cls

        self.state = TrackState.UNCONFIRMED

    def predict(self, kf: KalmanFilter):
        self.state_mean, self.state_covariance = kf.predict(
            self.state_mean, self.state_covariance)

        self.time_since_update += 1

    def update(self, kf: KalmanFilter, detection: Detection):
        self.state_mean, self.state_covariance = kf.update(self.state_mean,
                                                           self.state_covariance,
                                                           from_xyxy_to_xyah(detection.coords))

        self.hits += 1
        self.time_since_update = 0

        if self.state == TrackState.UNCONFIRMED and self.hits >= self.hits_until_confirm:
            self.state = TrackState.CONFIRMED

    def no_match(self):

        if self.state == TrackState.UNCONFIRMED:
            self.state = TrackState.DELETED
            return True
        elif self.time_since_update > self.max_since_update:
            self.state = TrackState.DELETED
            return True

        return False

    def get_position(self):
        center_x, center_y, aspect_ratio, height = self.state_mean[:4]

        width = aspect_ratio * height
        start_x = (center_x - (width // 2)).item()
        end_x = (start_x + width).item()
        start_y = (center_y - (height // 2)).item()
        end_y = (start_y + height).item()

        state_coords = CoordinatesXYXY(start_x=start_x, end_x=end_x,
                                       start_y=start_y, end_y=end_y)

        return state_coords





from typing import List, NewType, Tuple
from pydantic import BaseModel, ConfigDict
from deep_sort_reid.enums.tracker import TrackState
from deep_sort_reid.types.coords import CoordinatesXYXY

TrackID = NewType('TrackID', int)

class TrackResult(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    track_id: TrackID
    cls: List[int]  # Classes separated by commas
    frame_idx: int
    coords: CoordinatesXYXY
    state: TrackState
    color: Tuple





from typing import List, Sequence, Tuple
from deep_sort_reid.metric.IouMetric import IouMetric
from deep_sort_reid.metric.GatedMetric import GatedMetric
from deep_sort_reid.tracker.Track import Track
from deep_sort_reid.types.detection import Detection
from scipy.optimize import linear_sum_assignment


def min_cost_matching(metric: GatedMetric | IouMetric,
                      tracks: List[Track],
                      detections: List[Detection],
                      tracks_idx: List[int],
                      detections_idx: List[int]) -> Tuple[List[Tuple[int, int]], List[int], List[int]]:

    if len(detections_idx) == 0 or len(tracks_idx) == 0:
        return [], tracks_idx, detections_idx

    selected_tracks: List[Track] = []
    selected_detections: List[Detection] = []

    for track_idx in tracks_idx:
        selected_tracks.append(tracks[track_idx])

    for detection_idx in detections_idx:
        selected_detections.append(detections[detection_idx])

    cost_matrix = metric(selected_tracks, selected_detections)

    assert cost_matrix is not None

    cost_matrix[cost_matrix > metric.max_distance] = metric.max_distance + 1e-5

    (rows_idx, cols_idx) = linear_sum_assignment(cost_matrix)

    matches_track_det_idx: List[Tuple[int, int]] = []
    unmatched_tracks_idx: List[int] = []
    unmatched_detections_idx: List[int] = []

    for col_idx, detection_idx in enumerate(detections_idx):
        if col_idx not in cols_idx:
            unmatched_detections_idx.append(detection_idx)

    for row_idx, track_idx in enumerate(tracks_idx):
        if row_idx not in rows_idx:
            unmatched_tracks_idx.append(track_idx)

    for idx, row_idx in enumerate(rows_idx):
        track_idx = tracks_idx[row_idx]

        # the column that matches with the row in the assignment
        col_match = cols_idx[idx]
        detection_idx = detections_idx[col_match]

        if cost_matrix[row_idx, col_match] > metric.max_distance:
            unmatched_tracks_idx.append(track_idx)
            unmatched_detections_idx.append(detection_idx)
        else:
            matches_track_det_idx.append((track_idx, detection_idx))

    return matches_track_det_idx, unmatched_tracks_idx, unmatched_detections_idx



from setuptools import setup, find_packages

# Read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


def parse_requirements(filename):
    with open(filename, 'r') as req_file:
        return req_file.read().splitlines()


setup(
    name="deep_sort_reid",  # Your project name
    version="0.1.4",  # Your project version
    description="A re-mastered version of the original Deep Sort implementation, with added functionalities such as re-identification.",
    long_description=long_description,
    long_description_content_type="text/markdown",  # If you have README.md
    url="https://github.com/cajmorgan/deep_sort_reid",  # URL of your project repo
    author="Caj Morgan",
    author_email="caj@maiosolutions.com",
    license="GPL-3.0",  # License type
    packages=find_packages(),  # Automatically find and include your packages
    install_requires=[  # Dependencies (from your requirements.txt)
        *parse_requirements("requirements.txt")
    ],
    python_requires=">=3.10.0",  # Minimum Python version
)
